#!/usr/bin/env python

import asyncio
import argparse
import json
import numpy as np
import onnxruntime as ort
import sys
from tqdm.asyncio import tqdm

SAMPLE_RATE = 32000


def seconds_to_hms(seconds):
    hours, remainder = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(remainder, 60)
    return f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}"


def subsample(frame: np.ndarray, scale_factor: int) -> np.ndarray:
    subframe = frame[: len(frame) - (len(frame) % scale_factor)].reshape(
        -1, scale_factor
    )
    subframe_mean = subframe.max(axis=1)

    subsample = subframe_mean

    if len(frame) % scale_factor != 0:
        residual_frame = frame[len(frame) - (len(frame) % scale_factor) :]
        residual_mean = residual_frame.max()
        subsample = np.append(subsample, residual_mean)

    return subsample


def get_segments(
    scores: np.ndarray, precision: int, threshold: float, offset: int
) -> []:
    seq_iter = iter(np.where(scores > threshold)[0])
    try:
        seq = next(seq_iter)
        pred = scores[seq]
        segment = {"start": seq, "end": seq, "pred": pred}
    except StopIteration:
        return

    for seq in seq_iter:
        pred = scores[seq]
        if seq - 1 == segment["end"]:
            segment["end"] = seq
            segment["pred"] = max(segment["pred"], pred)
        else:
            segment["start"] = segment["start"]
            segment["end"] = segment["end"]

            yield segment
            segment = {"start": seq, "end": seq, "pred": pred}

    yield segment


def compute_timestamps(
    framewise_output: np.ndarray,
    precision: int,
    threshold: float,
    focus_idx: int,
    offset: int,
) -> []:
    focus = framewise_output[:, focus_idx]
    # precision in the amount of milliseconds per timestamp sample (higher values will result in less precise timestamps)

    subsampled_scores = subsample(focus, precision)
    segments = map(
        lambda segment: {
            "start": int(segment["start"] * precision / 100 + offset),
            "end": int(segment["end"] * precision / 100 + offset + precision / 100),
            "pred": round(float(segment["pred"]), 6),
        },
        get_segments(subsampled_scores, precision, threshold, offset),
    )
    return segments


def pad_array_if_needed(arr, desired_size, pad_value=0):
    current_size = arr.shape[0]
    if current_size < desired_size:
        padding_needed = desired_size - current_size
        padded_array = np.pad(
            arr, (0, padding_needed), "constant", constant_values=(pad_value,)
        )
        return padded_array
    else:
        return arr


async def load_audio(file: str, sr: int, frame_count: int) -> bytes:
    cmd = [
        "ffmpeg",
        "-hide_banner",
        "-loglevel",
        "warning",
        "-i",
        file,
        "-vn",
        "-af",
        "aresample=async=1:first_pts=0",
        "-ac",
        "1",
        "-f",
        "s16le",
        "-acodec",
        "pcm_s16le",
        "-ar",
        str(sr),
        "-",
    ]

    ffmpeg_process = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE
    )
    chunk_size = frame_count * 2

    while True:
        try:
            yield await ffmpeg_process.stdout.readexactly(chunk_size)
        except asyncio.IncompleteReadError as e:
            if e.partial:
                yield e.partial
            else:
                break

    await ffmpeg_process.wait()


async def process_video(
    file: str, model_dtype: np.dtype, args, ort_session, block_json: bool
):
    offset = 0
    frame_count = SAMPLE_RATE * args.block_size
    blocks = load_audio(file, SAMPLE_RATE, SAMPLE_RATE * args.block_size)

    info = {"filename": file, "timestamps": []}

    async for block in tqdm(blocks, file=sys.stderr, desc=f"Inferencing {file}"):
        samples = np.frombuffer(block, dtype=np.int16)
        samples = pad_array_if_needed(samples, frame_count)
        samples = samples.reshape(1, -1)
        samples = samples / (2**15)
        samples = samples.astype(model_dtype)

        ort_inputs = {"input": samples}
        framewise_output = ort_session.run(["output"], ort_inputs)[0]

        preds = framewise_output[0]
        info["timestamps"].extend(
            compute_timestamps(
                preds, args.precision, args.threshold, args.focus_idx, offset
            )
        )

        offset += args.block_size

    if not block_json:
        json.dump(info, sys.stdout)
        return None
    else:
        return info


async def main():
    parser = argparse.ArgumentParser(
        prog="bdetector", description="Scans audio files for sounds"
    )
    parser.add_argument(
        "files", metavar="f", nargs="+", type=str, help="files to be processed"
    )

    parser.add_argument(
        "--precision",
        metavar="p",
        nargs="?",
        type=int,
        default=1 * 100,
        help="the precision (in ms) of the timestamp selection process (higher is less precise)",
    )
    parser.add_argument(
        "--block_size",
        metavar="b",
        nargs="?",
        type=int,
        default=60 * 10,
        help="the block size is the amount of seconds (of samples) to process at once. Larger sizes offer better performance, but will consume significantly more memory",
    )
    parser.add_argument(
        "--threshold",
        metavar="t",
        nargs="?",
        type=float,
        default=0.90,
        help="The confidence threshold for a sound to be reported from 0-1",
    )
    parser.add_argument(
        "--focus_idx",
        metavar="i",
        nargs="?",
        type=int,
        help="the index of the audio class to track",
        default=58,
    )
    parser.add_argument(
        "--model",
        metavar="m",
        type=str,
        help="the location of the model file",
        required=True,
    )
    parser.add_argument(
        "--block_json",
        default=False,
        type=bool,
        help="Determines if the output should be sent when execution finishes, and inside a valid JSON container- or whether lines of individual JSON should be output",
    )
    parser.add_argument(
        "-N",
        default=1,
        type=int,
        help="The number of parallel decode tasks to perform. Increase this only if you are unable to saturate the processor/are limited by individual file access speed",
    )

    args = parser.parse_args()

    sess_options = ort.SessionOptions()
    sess_options.graph_optimization_level = ort.GraphOptimizationLevel.ORT_ENABLE_ALL
    sess_options.optimized_model_filepath = args.model

    providers = ort.get_available_providers()
    # TensorRT causes issues on Colab, you may remove the following lines to reenable TensorRT as an available provider
    # if 'TensorrtExecutionProvider' in providers:
    #     providers.remove('TensorrtExecutionProvider')

    ort_session = ort.InferenceSession(args.model, sess_options, providers=providers)

    stream_results = [
        process_video(file, np.float32, args, ort_session, args.block_json)
        for file in args.files
    ]

    semaphore = asyncio.Semaphore(args.N)

    async def sem_task(task):
        async with semaphore:
            return await task

    infos = await asyncio.gather(*(sem_task(task) for task in stream_results))

    if args.block_json:
        json.dump(infos, sys.stdout)


if __name__ == "__main__":
    asyncio.run(main())
